//! This module contains API definitions which are not considered
//! to be part diesels public API
//!
//! **DO NOT EXPECT ANY STABILITY GUARANTEES HERE**

pub mod alias_macro;
pub mod derives;
pub mod operators_macro;
pub mod table_macro;
